#!/usr/bin/python3
# --
# Copyright (C) 2024 it-novum GmbH, http://www.it-novum.com
# --
# Authors: Markus Stollwerk, Jens Michelsons
# --
# License ...
# --

from suds.client import Client
from pprint import pprint
from sys import exit
import argparse
import json
import re
import os
import time
import sys

PROCESS_STATE = {
    'SAPControl-J2EE-RUNNING': 'running',
    'SAPControl-J2EE-STOPPED': 'stopped',
    'SAPControl-J2EE-STARTING': 'starting',
    'SAPControl-J2EE-STOPPING': 'stopping',
    'SAPControl-J2EE-MAINTENANCE': 'maintenance',
    'SAPControl-J2EE-UNKNOWN': 'unknown',
    'SAPControl-J2EE-CORE-RUNNING': 'core-running'
}

SAP_STATE = {
    'SAPControl-GREEN': 0,
    'SAPControl-YELLOW': 1,
    'SAPControl-RED': 2,
    'SAPControl-GRAY': 2
}

NAGIOS_STATE_NAME = {
    0: "OK",
    1: "WARNING",
    2: "CRITICAL",
    3: "UNKNOWN"
}

LOG_LAST_ERROR = 15

parser = argparse.ArgumentParser()
optional = parser._action_groups.pop()
required = parser.add_argument_group('required arguments')
required.add_argument('--hostname', help='hostnamae or ipadresse', required=True)
required.add_argument('--port', help='SAP Webservice Port 5XX13', required=True)
required.add_argument('--mode', help='[checkitem, gettree, workprocess, checksyslog, getsyslog, j2ee-processlist, processlist, readlogfile, readsyslogfile, wpcountuser]', required=True)
optional.add_argument('--username', help='[username to check with]')
optional.add_argument('--password', help='[password to check with]')
optional.add_argument('--mode2', help='[list, invertthresholds, regex]')
optional.add_argument('--warning', help='warning threshold')
optional.add_argument('--critical', help='critical threshold')
optional.add_argument('--disablerfc', action='store_true', help='add this parameter to write disable-file for RFC-execution wrapper')
optional.add_argument('--item', help='item to check')
optional.add_argument('--logname', help='logname to check')
optional.add_argument('--logpattern', help='pattern to check')
optional.add_argument('--logidentifier', help='unique identifier if analyze more patterns in same log')
optional.add_argument('--loglasterror', help='plugin returns OK if no new error has been detected in the last n minutes')
optional.add_argument('--processokstates', help='Ok states can be a comma separated list out of [running, stopped, starting, stopping, maintenance, unknown, core-running]. Default: running')
optional.add_argument('--statearg', help='Parameter to store last check timestamp per syslog check')
parser._action_groups.append(optional) # added this line
args=parser.parse_args()

hostname=args.hostname
port=args.port
mode=args.mode
mode2=args.mode2
warning=args.warning
critical=args.critical
item=args.item
username=args.username
password=args.password
logname=args.logname
logpattern=args.logpattern
logidentifier = args.logidentifier
loglasterror = args.loglasterror
processokstates = args.processokstates
statesyslog=args.statearg

RC=0
RCtxt=""
RCok=0
RCwarning=1
RCcritical=2
RCunknown=3
maxage=900
delimiter="##"

url="http://" + hostname  + ":" + port  + "/?wsdl"
tmpfile="/var/tmp/check_sap_webservices/" + hostname + "-" + port + ".tmp"
tmpfilesyslog="/var/tmp/check_sap_webservices/" + hostname + "-" + port + ".syslog.tmp"

if mode == "checksyslog":
    try:
        lastsyslog="/var/tmp/check_sap_webservices/" + hostname + "-" + port + ".syslog." + statesyslog + ".time.tmp"
    except:
        print("UNKNOWN- please provide statearg")
        exit(RCunknown)



if not os.path.exists("/var/tmp/check_sap_webservices/"):
    os.makedirs("/var/tmp/check_sap_webservices/")


tmpitems={}
resultitems={}

def GetAlertTree():
    try:
        client = Client(url, cache=None)
        result = client.service.GetAlertTree()
        result = result[0]
    except Exception as e:
        print("CRITICAL - could not connect to host %s on port %s: %s" % (hostname, port, str(e)))
        exit(RCcritical)

    if len(result) == 0:
        print("CRITICAL - could not query data")
        exit(RCcritical)

    for counter in range(len(result)):
        tmpdict={}
        if result[counter].name is None:
            tmpdict["name"]="SAP"
        else:
            tmpdict["name"]=result[counter].name

        tmpdict["parent"]=result[counter].parent
        tmpitems[counter]=tmpdict

        path=""
        parent=result[counter].parent
        while parent > -1:
            path=tmpitems[parent]["name"] + delimiter + path
            parent=tmpitems[parent]["parent"]
        if result[counter].name is None:
            path=path + "SAP"
        else:
            path=path + result[counter].name

        tmpresult={}
        tmpresult["state"]=result[counter].ActualValue
        tmpresult["output"]=result[counter].description
        tmpresult["tid"]=result[counter].Tid
        resultitems[path]=tmpresult

    if mode2 == "list":
        pprint(resultitems)
    else:
        with open(tmpfile, 'w') as file:
            file.write(json.dumps(resultitems))

    print("OK - data receive")
    exit(RCok)

def ABAPReadSyslog():
    try:
        client = Client(url, cache=None)
        result = client.service.ABAPReadSyslog()
        result = result[0]
    except:
        print("CRITICAL - could not connect to host %s on port %s" % (hostname, port))
        exit(RCcritical)

    if len(result) == 0:
        print("CRITICAL - could not query data")
        exit(RCcritical)


    for counter in range(len(result)):

        tmpresult={}
        tmpresult["time"]=time.mktime(time.strptime(result[counter].Time, "%Y %m %d %H:%M:%S"))
        tmpresult["user"]=result[counter].User
        tmpresult["client"]=result[counter].Client
        tmpresult["text"]=result[counter].Text
        tmpresult["state"]=result[counter].Severity
        resultitems[counter]=tmpresult

    with open(tmpfilesyslog, 'w') as file:
        file.write(json.dumps(resultitems))

    print("OK - data receive")
    exit(RCok)

def GetAlerts(tid):
    try:
        client = Client(url, cache=None)
        result = client.service.GetAlerts(tid)
        result=result[1][0]
        longoutput=""
        for counter in range(len(result)):
            longoutput=longoutput + result[counter].Time + " --- " + result[counter].Description + "\n"
    except:
        longoutput=""

    return longoutput

def GetValue(argitem):
    try:
        with open(tmpfile, 'r') as file:
            resultitems=json.loads(file.read())
    except:
        print("UNKNOWN - could not open tmpfile or its empty " + tmpfile)
        exit(RCunknown)

    if (time.time() - os.stat(tmpfile).st_mtime) > maxage:
        print("UNKNOWN - last data receive is to old %ds > %ds" % ((time.time() - os.stat(tmpfile).st_mtime), maxage))
        exit(RCunknown)
    for key in resultitems:
        if key.endswith(argitem):
            return resultitems[key]

def GetSyslogValue():

    returnvalues={}
    counttmp=0
    lasttime=""
    try:
        with open(tmpfilesyslog, 'r') as file:
            resultitems=json.loads(file.read())
    except:
        print("UNKNOWN - could not open tmpfile or its empty " + tmpfilesyslog)
        exit(RCunknown)

    if (time.time() - os.stat(tmpfilesyslog).st_mtime) > maxage:
        print("UNKNOWN - last data receive is to old %ds > %ds" % ((time.time() - os.stat(tmpfilesyslog).st_mtime), maxage))
        exit(RCunknown)


    try:
        with open(lastsyslog, 'r') as file:
            lasttime=float(file.read())
            if loglasterror:
                lasttime = lasttime - float(loglasterror) * 60
    except:
        with open(lastsyslog, 'w') as file:
            file.write(str(time.time()))
            return None


    for counter in range(len(resultitems)):
        if resultitems[str(counter)]["time"] > lasttime:
            returnvalues[counttmp]=resultitems[str(counter)]
            counttmp=counttmp+1

    with open(lastsyslog, 'w') as file:
        file.write(str(resultitems["0"]["time"]))
    return(returnvalues)

def ABAPGetWPTable(argitem):

    try:
        client = Client(url, cache=None)
        result = client.service.ABAPGetWPTable()
        result = result[0]
    except:
        print("CRITICAL - could not connect to host %s on port %s" % (hostname, port))
        exit(RCcritical)

    if len(result) == 0:
        print("CRITICAL - could not query data")
        exit(RCcritical)

    count=0
    for counter in range(len(result)):
            if str(result[counter].Typ) == argitem and result[counter].Status == "Wait":
                count=count+1

    return count

def ABAPWPTableUsers():

    try:
        client = Client(url, cache=None)
        result = client.service.ABAPGetWPTable()
        result = result[0]
    except:
        print("CRITICAL - could not connect to host %s on port %s" % (hostname, port))
        exit(RCcritical)

    if len(result) == 0:
        print("CRITICAL - could not query data")
        exit(RCcritical)

    count=0

    users_table = {}
    for counter in range(len(result)):
        if result[counter].User:
            if result[counter].User not in users_table:
                users_table[result[counter].User] = 1
            else:
                users_table[result[counter].User] += 1

    print("OK - Check to display workprocess utilization by all users")

    if len(users_table) > 0:
        for user, count in users_table.items():
            print(f'{user}: {count}')
    else:
        print("No active user sessions")

def WPTableCountUsers(argitem):

    try:
        client = Client(url, cache=None)
        result = client.service.ABAPGetWPTable()
        result = result[0]
    except:
        print("CRITICAL - could not connect to host %s on port %s" % (hostname, port))
        exit(RCcritical)

    if len(result) == 0:
        print("CRITICAL - could not query data")
        exit(RCcritical)

    count=0
    for counter in range(len(result)):
            if str(result[counter].User) == argitem:
                count=count+1

    return count

def ReadSyslogFile(logfilter):
    count_matches = 0
    startcookie = 0
    last_error = None
    lookback = LOG_LAST_ERROR

    if loglasterror:
        lookback = int(loglasterror)

    max_error_age = time.localtime(time.time() - lookback * 60)

    if logfilter == None:
        logfilter = "*"

    Filter = 'Severity#Time)%s#Type#Client#User#Tcode#MNo#Pid#Terminal#Program#Session#Text=*%s*' % (time.strftime("%Y %m %d %H:%M:%S", max_error_age), logfilter)
    client = Client(url, cache=None)

    try:
        Logfiles = client.service.ListLogFiles()
        for Logfile in Logfiles[0]:
            if Logfile['format'] == 'SAP Syslog':
                logname = Logfile['filename']

        result = client.service.ReadLogFile(maxentries = 0, filename = logname, filter = Filter)
    except:
       print("CRITICAL - could not connect to syshost %s on port %s" % (hostname, port))
       exit(RCcritical)

    jsoninfo = {}
    if result['fields']:
        count_matches = len(result['fields'][0])

    if count_matches > 0:
        print("CRITICAL - Pattern '%s' in logfile '%s' found %s times in last %s minutes" % (logfilter, logname, count_matches, lookback))
        exit(RCcritical)
    else:
        print("OK - Pattern '%s' in logfile '%s' not found in last %s minutes" % (logfilter, logname, lookback))
        exit(RCok)


def ReadLogFile(logname, logfilter):
    count_matches = 0
    startcookie = 0
    last_error = None
    max_error_age = LOG_LAST_ERROR

    if loglasterror:
        max_error_age = int(loglasterror)

    tmpfilelog = "/var/tmp/check_sap_webservices/" + hostname + "-" + port + "-" + logname.replace('/', '_') + "-log"
    if logidentifier:
        tmpfilelog = tmpfilelog + "." + logidentifier
    tmpfilelog = tmpfilelog + ".tmp"

    if (os.path.isfile(tmpfilelog)):
        try:
            tmpfile_in = open(tmpfilelog)
            data = json.load(tmpfile_in)
            startcookie = data['endcookie']
            last_error = data['lasterror']
            tmpfile_in.close()
        except:
            print("Critical - Error while trying to access data.")
            sys.exit(RCcritical)

    client = Client(url, cache=None)

    try:
        result = client.service.ReadLogFile(statecookie = startcookie, maxentries = "0", filename = logname, filter="CASEIGNORE:Line=*" + logfilter + "*")
    except:
        startcookie = 0

    if startcookie == 0:
        try:
            result = client.service.ReadLogFile(statecookie = startcookie, maxentries = "0", filename = logname, filter="CASEIGNORE:Line=*" + logfilter + "*")
        except:
            print("CRITICAL - could not connect to host %s on port %s" % (hostname, port))
            exit(RCcritical)

    jsoninfo = {}
    matches_found = len(result[3])
    endcookie = result[2]

    if matches_found > 0:
        count_matches = len(result[3][0])
        last_error = time.time()

    jsoninfo['lasterror'] = last_error

    if int(startcookie) > int(endcookie):
        jsoninfo['endcookie'] = 0
    else:
        jsoninfo['endcookie'] = endcookie

    try:
        tmpfile_out = open(tmpfilelog, "w")
        json.dump(jsoninfo, tmpfile_out)
        tmpfile_out.close()
    except:
        print("CRITICAL - File is not writeable")
        exit(RCcritical)

    if last_error:
        minutes_since_last_error = int(time.time() - last_error) / 60
        if minutes_since_last_error < max_error_age:
            print("CRITICAL - Pattern '%s' in logfile '%s' found in last %s minutes" % (logfilter, logname, max_error_age))
            exit(RCcritical)

    print("OK - Pattern '%s' in logfile '%s' not found in last %s minutes" % (logfilter, logname, max_error_age))
    exit(RCok)


def J2EEGetProcessList(argitem):
    processlist = {}
    client = Client(url, cache=None)
    ok_list = ['running']
    exit_ok = 0
    if processokstates:
        ok_list = processokstates.split(',')

    try:
        result = client.service.J2EEGetProcessList()
        result = result[0]
    except:
        print("CRITICAL - could not connect to host %s on port %s" % (hostname, port))
        exit(RCcritical)

    for item in result:
        processlist[item['name']] = item

    try:
        if PROCESS_STATE[processlist[argitem]['state']] in ok_list:
            print("OK - %s (%s) is %s" % (argitem, processlist[argitem]['type'], PROCESS_STATE[processlist[argitem]['state']]))
            exit_ok = 1
    except:
        print("UNKNOWN - Item %s does not exist." % argitem)
        exit(RCunknown)

    if exit_ok:
        exit(RCok)

    print("CRITICAL - %s (%s) is %s should be %s" % (argitem, processlist[argitem]['type'], PROCESS_STATE[processlist[argitem]['state']], ok_list))
    exit(RCok)

def GetProcessList(argitem):
    processlist = {}
    client = Client(url, cache=None)
    exit_code = -1
    try:
        result = client.service.GetProcessList()
        result = result[0]
    except:
        print("CRITICAL - could not connect to host %s on port %s" % (hostname, port))
        exit(RCcritical)

    for item in result:
        if item['name'] == argitem:
            exit_code = SAP_STATE[item['dispstatus']]

            print("%s - Process %s is %s" % (NAGIOS_STATE_NAME[exit_code], argitem, item['textstatus']))

    if exit_code == -1:
        print("UNKNOWN - Item not found")
        exit(3)
    exit(exit_code)

def NagiosOut(argvalues):

    if mode == "checkitem":
        if not delimiter in item:
            print("UNKNOWN - delimiter %s not found in item %s please submit at least one parent like parent##metric" % (delimiter, item))
            exit(RCunknown)
        vitem=item.replace(delimiter, " ")
        try:
            vstate=argvalues["state"]
            voutput=argvalues["output"]
            voutputlong=argvalues["output"]
            vtid=argvalues["tid"]
            if voutput is not None:
                try:
                    voutput=float(re.findall('[0-9]+', voutput)[0])
                except:
                    voutput=None
            else:
                voutputlong=vstate
        except:
            print("UNKNOWN - could not find " + item)
            exit(RCunknown)

        lout=GetAlerts(vtid)


        if warning is not None and critical is not None and warning != "na" and critical != "na":
            warn=float(re.findall('[0-9]+', warning)[0])
            crit=float(re.findall('[0-9]+', critical)[0])
            if mode2 == "invertthresholds":
                if warn < crit:
                    print("UNKNOWN - in mode invertthresholds warning should be larger then critical")
                    exit(RCunknown)

                if voutput > warn:
                    RCtxt="OK"
                    RC=RCok
                elif voutput < crit:
                    RCtxt="CRITICAL"
                    RC=RCcritical
                else:
                    RCtxt="WARNING"
                    RC=RCwarning
                print("%s - %s is %s|%s=%s" % (RCtxt, vitem, voutputlong, vitem.replace(' ', '_'), voutput))
            else:
                if warn > crit:
                    print("UNKNOWN - warning should be smaller then critical")
                    exit(RCunknown)

                if voutput < warn:
                    RCtxt="OK"
                    RC=RCok
                elif voutput > crit:
                    RCtxt="CRITICAL"
                    RC=RCcritical
                else:
                    RCtxt="WARNING"
                    RC=RCwarning
                print("%s - %s is %s|%s=%s;%s;%s" % (RCtxt, vitem, voutputlong, vitem.replace(' ', '_'), voutput, warn, crit))

        else:
            if "GREEN" in vstate:
                RCtxt="OK"
                RC=RCok
            elif "YELLOW" in vstate:
                RCtxt="WARNING"
                RC=RCwarning
            elif "RED" in vstate:
                RCtxt="CRITICAL"
                RC=RCcritical
            elif "GRAY" in vstate:
                RCtxt="UNKNOWN"
                RC=RCunknown
            print("%s - %s is %s|%s=%s" % (RCtxt, vitem, voutputlong, vitem.replace(' ', '_'), voutput))
        if len(lout) > 0:
            print(lout)
        exit(RC)

    elif mode == "workprocess":
        voutput=int(argvalues)
        try:
            warn=int(re.findall('[0-9]+', warning)[0])
            crit=int(re.findall('[0-9]+', critical)[0])
        except:
            print("UNKNOWN - please submit thresholds")
            exit(RCunknown)

        if crit > warn:
            print("UNKNOWN - critical threshold should be less then warning")

        if voutput > warn:
            RCtxt="OK"
            RC=RCok
        elif voutput <= crit:
            RCtxt="CRITICAL"
            RC=RCcritical
        else:
            RCtxt="WARNING"
            RC=RCwarning

        print("%s - there are %s %s workprocess in state Wait|%s=%s" % (RCtxt, voutput, item, item, voutput))
        exit(RC)

    elif mode == "wpcountuser":
        voutput=int(argvalues)
        try:
            warn=int(re.findall('[0-9]+', warning)[0])
            crit=int(re.findall('[0-9]+', critical)[0])
        except:
            print("UNKNOWN - please submit thresholds")
            exit(RCunknown)
        file_path = "/var/tmp/check_sap_webservices/rfc_execution/%s.disabled" % hostname
        if voutput > crit:
            if args.disablerfc:
                open(file_path, 'a').close()
            RCtxt="CRITICAL"
            RC=RCcritical
        elif voutput > warn:
            RCtxt="WARNING"
            RC=RCwarning
        else:
            if os.path.exists(file_path):
                os.remove(file_path)
            RCtxt="OK"
            RC=RCok

        print("%s - User %s uses %s workprocesses. | used_workprocesses=%s" % (RCtxt, item, voutput, voutput))
        exit(RC)

    elif mode == "checksyslog":
        errentries=""
        errentrieswarn=""
        errentriescrit=""
        RC=RCok
        if argvalues is None:
            print("OK - first run. Wait for the next check")
            exit(RCok)
        elif len(argvalues) == 0:
            print("OK - no new entries since last check")
            exit(RCok)
        else:
            if mode2 == "regex":
                if item is None:
                    print("UNKNOWN - please submit regex item to check")
                    exit(RCunknown)
                else:
                    for counter in range(len(argvalues)):
                        if len(re.findall(item, argvalues[counter]["text"], re.IGNORECASE)) > 0:
                            errentries=errentries+"Time: %s, State %s, Text: %s\n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(argvalues[counter]["time"])), argvalues[counter]["state"], argvalues[counter]["text"])

                    if errentries == "" :
                        print("OK - pattern %s does not match syslog entries" % (item))
                        RC=RCok
                    else:
                        print("CRITICAL - pattern %s matches syslog entries: %s ......." % (item, errentries[:4000].replace('\n',';')))
                        print(errentries)
                        RC=RCcritical

            else:
                for counter in range(len(argvalues)):
                    if argvalues[counter]["state"] == "SAPControl-YELLOW":
                        errentrieswarn=errentrieswarn+"Time: %s, State %s, Text: %s\n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(argvalues[counter]["time"])), argvalues[counter]["state"], argvalues[counter]["text"])
                    if argvalues[counter]["state"] == "SAPControl-RED":
                        errentriescrit=errentriescrit+"Time: %s, State %s, Text: %s\n" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(argvalues[counter]["time"])), argvalues[counter]["state"], argvalues[counter]["text"])

                errentries=errentriescrit+errentrieswarn

                if errentries == "":
                    print("OK - no warnings or errors found")
                    RC=RCok
                elif errentriescrit != "":
                    print("CRITICAL - critical errors found: %s ......." % (errentries[:4000].replace('\n',';')))
                    print(errentries)
                    RC=RCcritical
                else:
                    print("WARNING - warning errors found: %s ......." % (errentries[:4000].replace('\n',';')))
                    print(errentries)
                    RC=RCwarning


        exit(RC)


#Execute selected Mode
if mode == "gettree":
        GetAlertTree()
elif mode == "checkitem":
        if item is None:
                print("UNKNOWN - please submit item to check")
                exit(RCunknown)
        else:
                NagiosOut(GetValue(item))
elif mode == "workprocess":
    NagiosOut(ABAPGetWPTable(item))
elif mode == "wpcountuser":
    NagiosOut(WPTableCountUsers(item))
elif mode == "wpusers":
    ABAPWPTableUsers()
elif mode == "getsyslog":
    ABAPReadSyslog()
elif mode == "checksyslog":
    NagiosOut(GetSyslogValue())
elif mode == "readlogfile":
    ReadLogFile(logname, logpattern)
elif mode == "readsyslogfile":
    ReadSyslogFile(logpattern)
elif mode == "j2ee-processlist":
    J2EEGetProcessList(item)
elif mode == "processlist":
    GetProcessList(item)
else:
        print("UNKNOWN - unknown mode %s" % (mode))
        exit(RCunknown)
